/*
 * This file is a component of thundr, a software library from 3wks.
 * Read more: http://www.3wks.com.au/thundr
 * Copyright (C) 2013 3wks, <thundr@3wks.com.au>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.threewks.thundr.googleapis.prediction;

import static com.atomicleopard.expressive.Expressive.mapKeys;

import java.io.IOException;
import java.util.Arrays;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import com.google.api.client.googleapis.auth.oauth2.GoogleCredential;
import com.google.api.client.http.javanet.NetHttpTransport;
import com.google.api.client.json.gson.GsonFactory;
import com.google.api.services.prediction.Prediction;
import com.google.api.services.prediction.model.Input;
import com.google.api.services.prediction.model.Input.InputInput;
import com.google.api.services.prediction.model.Output;
import com.google.api.services.prediction.model.Output.OutputMulti;
import com.google.api.services.prediction.model.Training;

public class GooglePredictionServiceImpl implements GooglePredictionService {

	private static final Map<String, ModelStatus> statusMap = mapKeys("DONE", "RUNNING").to(ModelStatus.Ready, ModelStatus.Training);
	private Prediction prediction;

	public GooglePredictionServiceImpl(GoogleCredential googlePredictionCredential) {
		this.prediction = new Prediction.Builder(new NetHttpTransport(), new GsonFactory(), googlePredictionCredential).setApplicationName("thundr-google-prediction/1.0").build();
	}

	@Override
	public Prediction getPrediction() {
		return prediction;
	}

	@Override
	public void trainModel(String modelName, String modelLocation) {
		Training training = new Training();
		training.setId(modelName);
		training.setStorageDataLocation(modelLocation);
		try {
			prediction.trainedmodels().insert(training).execute();
		} catch (IOException e) {
			throw new PredictionException(e, "Failed to train model '%s': %s", modelName, e.getMessage());
		}
	}

	@Override
	public ModelStatus getModelStatus(String modelName) {
		try {
			String status = prediction.trainedmodels().get(modelName).execute().getTrainingStatus();
			ModelStatus modelStatus = statusMap.get(status);
			return modelStatus == null ? ModelStatus.Unknown : modelStatus;
		} catch (IOException e) {
			throw new PredictionException(e, "Failed to train model '%s': %s", modelName, e.getMessage());
		}
	}

	@Override
	public String predictClassification(String modelName, Object... fields) {
		Output output = predict(modelName, fields);
		return output.getOutputLabel();
	}

	@Override
	public Map<String, Double> predictClassifications(String modelName, Object... fields) {
		Map<String, Double> result = new LinkedHashMap<String, Double>();
		Output output = predict(modelName, fields);
		List<OutputMulti> outputMulti = output.getOutputMulti();
		for (OutputMulti weight : outputMulti) {
			result.put(weight.getLabel(), weight.getScore());
		}

		return result;
	}

	@Override
	public Output predict(String modelName, Object... fields) {
		Output output;
		Input input = new Input();
		InputInput inputInput = new InputInput();
		inputInput.setCsvInstance(Arrays.asList(fields));
		input.setInput(inputInput);
		try {
			output = prediction.trainedmodels().predict(modelName, input).execute();
		} catch (IOException e) {
			throw new PredictionException(e, "Failed to predict with model '%s': %s", modelName, e.getMessage());
		}
		return output;
	}
}
